/* $Id$ */
/*

	24LC64 Serial EEPROM Driver 

        Copyright (C) 1998 Koji Suzuki

        24LC64 Υɥ饤СեΤ褦 read/write 뤳ȤǤ롣
        ưǽĤ


        SCL --- Port2 - 4
        SDA --- Port2 - 5
        (ץ륢å MOS Ѥʤ(Ǥʤ)硢10K 餤ǥץ륢å
         뤳ȡ)

        A0,A1,A2,WP ϡVSS ³


        ǽ
          eeprom_init() --- 24LC64 ѤǤ褦ˤȤȤˡ
                            ˽Ƥ뤫ɤ(IDENT 񤭹ޤ
                            뤫ɤ)åɬפʤ饼ꥢԤ

                            ޤץ ROM Υåꡢ
                            ROM äƤСeeprom Υ󥿤
                            +1 롣

         eeprom_read(offset,addr,bytes)
                            EEPROM  offset ΰ addr Ϥޤ
                            žｪλ 0 ֤

         eeprom_write(offset,addr,bytes)
                            EEPROM  offset ΰ addr Ϥޤ
                            Ƥžｪλ 0 ֤



*/


#include "common.h"

#define eeprom_ddr	p2_ddr
#define EEPROM_DDR	P2DDR
#define EEPROM_DR	P2DR
#define EEPROM_PCR	P2PCR
#define SCL		(1<<4)
#define SDA		(1<<5)


static write_bit(dat) {
	eeprom_ddr |= SCL;
	outb(EEPROM_DDR,eeprom_ddr);
	if (dat) {
		eeprom_ddr &= ~SDA;
	} else {
		eeprom_ddr |= SDA;
	}
	outb(EEPROM_DDR,eeprom_ddr);
	nop();
	eeprom_ddr &= ~SCL;
	outb(EEPROM_DDR,eeprom_ddr);
	nop();
}

static read_bit() {
	int ret;
	eeprom_ddr &= ~SDA;
	eeprom_ddr |= SCL;
	outb(EEPROM_DDR,eeprom_ddr);
	nop();
	eeprom_ddr &= ~SCL;
	outb(EEPROM_DDR,eeprom_ddr);
	ret = inb(EEPROM_DR) & SDA;
	nop();
	return ret;
}

static start_condition() {
	eeprom_ddr &= ~SCL;
	outb(EEPROM_DDR,eeprom_ddr);
	nop();
	eeprom_ddr |= SDA;
	outb(EEPROM_DDR,eeprom_ddr);
	nop();
}

static stop_condition() {
	write_bit(0);
	nop();
	eeprom_ddr &= ~SDA;
	outb(EEPROM_DDR,eeprom_ddr);
	nop();
}

static write_byte(dat) int dat; {
	int i;
	int x = 0x80;
	for (i=0; i<8; i++) {
		write_bit(dat & x);
		x >>= 1;
	}
}

static read_byte() {
	int i;
	int ret=0;
	for (i=0; i<8; i++) { 
		ret <<= 1;
		ret |= read_bit()?1:0;
	}
	return ret;
}

struct eeprom_header {
	char ident[4];
	int sum;
	int fchg;
};

eeprom_init() {
	extern etext();
	char *p;
	int sum = 0;
	int ret;
	struct eeprom_header h;
#ifdef EEPROM_PCR
	orb(EEPROM_PCR,SCL|SDA);
#endif
	eeprom_ddr &= ~(SCL|SDA);
	outb(EEPROM_DDR,eeprom_ddr);

	ret = eeprom_read(0,&h,sizeof(h));
	if (ret) ret = eeprom_read(0,&h,sizeof(h));
	if (ret) return;
	p = 0;
	while (p < (char *)etext) {
		sum += *(p++);
	}
	if (h.ident[0] != 'K'
	 || h.ident[1] != 'M'
	 || h.ident[2] != 'D'
	 || h.ident[3] != 'S' ) {
		eeprom_clear();
		h.ident[0] = 'K';
		h.ident[1] = 'M';
		h.ident[2] = 'D';
		h.ident[3] = 'S';
		h.sum = sum;
		h.fchg = 0;
		ret = eeprom_write(0,&h,sizeof(h));
		return;
	}
	if (h.sum != sum) {
		h.fchg++;
		ret = eeprom_write(0,&h,sizeof(h));
	}
}

eeprom_clear() {
	char buf[32];
	int offset;
	int i,ret;
	for (i=0; i<32; i++) {
		buf[i] = 0;
	}
	for (offset = 0; offset < 0x2000; offset += 32) {
		if (ret =eeprom_write(offset,buf,32)) break;
	}
}

eeprom_write(offset,buf,bytes) int offset; char *buf; int bytes; {
	int ret = 0;
	int i;
	int f=1;
	while (bytes > 0) {
retry:
		start_condition();
		write_byte(0xa0);
		if (read_bit()) {
			if(!f)goto retry;
			ret = 1;
			goto err;
		}
		write_byte(offset >> 8);
		if (read_bit()) { ret = 2; goto err;}
		write_byte(offset);
		if (read_bit()) { ret = 3; goto err;}
		for (i=0; (i<32) && (bytes > 0) ; i++) {
			write_byte(*buf);
			if (read_bit()) { ret = 4; goto err;}
			offset++;
			buf++;
			bytes--;
		}
		stop_condition();
		f=0;
	}
	while (eeprom_read(0,&i,sizeof(i)))
		;
	return ret;
err:
	stop_condition();
	return ret;
}


eeprom_read(offset,buf,bytes) int offset; char *buf; int bytes; {
	int ret = 0;

	start_condition();
	write_byte(0xa0);
	if (read_bit()) { ret = 1; goto err; }
	write_byte(offset >> 8);
	if (read_bit()) { ret = 2; goto err; }
	write_byte(offset);
	if (read_bit()) { ret = 3; goto err; }
	stop_condition();
	
	start_condition();
	write_byte(0xa1);
	if (read_bit()) { ret = 4; goto err; }

	while (bytes > 0) {
		*buf = read_byte();
		if (bytes > 1)
			write_bit(0); /* ACK */
		else
			write_bit(1); /* NO ACK */
		offset++;
		buf++;
		bytes--;
	}
err:
	stop_condition();
	return ret;
}

